import numpy as np
import scipy.sparse as sp
from scipy.sparse.linalg import spsolve
from scipy.sparse import spmatrix
import matplotlib.pyplot as plt
import matplotlib.animation as animation

import mesh
from darcy_1D import getmatrhs1D


def time_loop(mesh, K, bcfunc, rhsfun, t0, tf, dt, theta, uinit, lineaire=True):
    t = t0
    ts = [t0]
    u = [uinit]
    uold = uinit
    while(t < tf):
        if (1.1*dt >= tf -t):
            dt = tf -t
        
        lval = theta*bcfunc(t+dt)[0]['value'] + (1-theta)*bcfunc(t)[0]['value'] 
        rval = theta*bcfunc(t+dt)[1]['value'] + (1-theta)*bcfunc(t)[1]['value']
        bcleft = {'type': bcfunc(t)[0]['type'], 'value': lval}
        bcright = {'type': bcfunc(t)[1]['type'], 'value': rval}
        bcs = [bcleft, bcright]
        if lineaire:
            unew = do_step(mesh, K, bcs, rhsfun(t+dt), dt, theta, uold)
        else:
            unew = picard(mesh, K, bcs, rhsfun(t+dt), dt, theta, uold)
        
        u.append(unew)
        t += dt
        ts.append(t)
        uold = unew

    return ts, u

def do_step(mesh, K, bclist, rhsfun, dt, theta, uold):

    A, f = getmatrhs1D(mesh, K, bclist, rhsfun)
    if theta == 0:
        unew = uold + (dt*f -dt*A.dot(uold))/mesh.mes
    else:
        b = 1/theta * (f + mesh.mes*uold/dt - (1-theta)*A.dot(uold))
        B = A.copy()
        D =A.diagonal()
        D += mesh.mes/(theta*dt)
        B.setdiag(D)
        unew = spsolve(B, b)
    return unew

def picard(mesh, Kfun, bclist, rhsfun, dt, theta, uold):
    converged = False
    tol =1e-5

    ukold = uold

    while not(converged):
        K = Kfun(ukold)
        A, f = getmatrhs1D(mesh, K, bclist, rhsfun)
        b = 1/theta * (f + mesh.mes*uold/dt - (1-theta)*A.dot(uold))
        B = A.copy()
        D =A.diagonal()
        D += mesh.mes/(theta*dt)
        B.setdiag(D)
        uknew = spsolve(B, b)
        
        critere = np.max (np.abs (uknew-ukold )) / np.max(np.abs(ukold))
        #print('Critere   ', critere)
        converged = critere < tol
        ukold = uknew

    return uknew
    

def test_heterogene():
    L = 1
    n = 100
    umesh = mesh.make_uniform_mesh(L, n)
    t0 = 0
    tf = 10000
    dt = 100
    theta = 1
    K = 1e-5*np.ones_like(umesh.centres)
    K[0:n//4] *= 1
    K[n//4:2*n//3] *= 10
    K[2*n//3:n] *= 100
    #K[0:n // 2] = 1e-4
    #K[n // 2:n] = 1e-5

    uinit = np.zeros_like(umesh.centres)
    uinit[0] = 50
    uinit[-1] = 50

    bcleft = {'type': 'Dirichlet', 'value': 100.}
    bcright = {'type': 'Dirichlet', 'value': 100}

    bclist = [bcleft, bcright]
    #   bclist = [bcleftD, bcrightN]
    #   bclist = [bcleftN, bcrightD]
    #   bclist = [bcleftN, bcrightN]

    def rhsfunc(t):
        return np.zeros_like(umesh.centres)

    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    line, = ax.plot([], [], 'o-', lw=2)

    ts, u = time_loop(umesh, K, bclist, rhsfunc, t0, tf, dt, theta, uinit)

    def animate(frame):
        ax.clear()
        plt.plot(umesh.centres, u[frame], label='Temp. calculee')
        plt.ylim([0, 100])
        plt.xlim([0, L])
        plt.legend()

    ani = animation.FuncAnimation(fig, animate, frames=len(u), interval=200,repeat=False)
    plt.show()

def test_homogene():
    L = 1
    n = 20
    h=L/n
    umesh = mesh.make_uniform_mesh(L, n)
    x = umesh.centres
    t0 = 0
    tf = .1
    dt = 0.01/4
    theta = 0.5
    K = np.ones_like(x)
    uinit = np.sin(np.pi*x/2) + 0.5 * np.sin(2*np.pi*x)

    def bcs(t):
        bcleft = {'type': 'Dirichlet', 'value': 0}
        bcright = {'type': 'Dirichlet', 'value': np.exp(-np.pi**2*t/4)}
        return [bcleft, bcright]

    def rhsfunc(t):
        return np.zeros_like(x)

    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    line, = ax.plot([], [], 'o-', lw=2)

    ts, u = time_loop(umesh, K, bcs, rhsfunc, t0, tf, dt, theta, uinit)

    def solex(x, t):
        return np.exp(-np.pi**2*t/4) * np. sin(np.pi*x/2.) \
               + 0.5 * np.exp(-4*np.pi**2*t) * np. sin(2*np.pi*x)

    print("Fourier= ", K*dt/h**2)
    errxt=0
    for i in range(len(u)):
        errxt += np.linalg.norm(u[i] - solex(x,ts[i]),2)**2*h
    errxt = np.sqrt(errxt*dt)
    errfin = np.linalg.norm(u[-1] - solex(x,ts[-1]),2)*np.sqrt(h)
    print("erreur x-t, finale= ", errxt, errfin)

       
    def animate(frame):
        ax.clear()
        plt.plot(x,u[frame], '-+', x,solex(x, ts[frame]), )
        plt.legend(['Temp. calculee', 'Temp exacte'])
        plt.ylim([0, 1])
        plt.xlim([0, L])

    plotit=True
    if plotit:
        ani = animation.FuncAnimation(fig, animate, frames=len(u), interval=100,repeat=False)
        plt.show()


    
if __name__ == '__main__':
    test_homogene()
